"""
UV Dataset loader
"""
import os
import logging
import time

import deepchem
from deepchem.molnet.load_function.kaggle_features import merck_descriptors
from deepchem.molnet.load_function.uv_tasks import UV_tasks
from deepchem.utils import remove_missing_entries

logger = logging.getLogger(__name__)

TRAIN_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/UV_training_disguised_combined_full.csv.gz"
VALID_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/UV_test1_disguised_combined_full.csv.gz"
TEST_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/UV_test2_disguised_combined_full.csv.gz"

TRAIN_FILENAME = "UV_training_disguised_combined_full.csv.gz"
VALID_FILENAME = "UV_test1_disguised_combined_full.csv.gz"
TEST_FILENAME = "UV_test2_disguised_combined_full.csv.gz"


def get_transformers(train_dataset):
    "Gets transformations applied on the dataset"

    transformers = list()

    return transformers


def gen_uv(UV_tasks, data_dir, train_dir, valid_dir, test_dir, shard_size=2000):
    """Loading the UV dataset; does not do train/test split"""

    time1 = time.time()

    train_files = os.path.join(data_dir, TRAIN_FILENAME)
    valid_files = os.path.join(data_dir, VALID_FILENAME)
    test_files = os.path.join(data_dir, TEST_FILENAME)

    # Download files if they don't exist

    if not os.path.exists(train_files):

        logger.info("Downloading training file...")
        deepchem.utils.data_utils.download_url(url=TRAIN_URL, dest_dir=data_dir)
        logger.info("Training file download complete.")

        logger.info("Downloading validation file...")
        deepchem.utils.data_utils.download_url(url=VALID_URL, dest_dir=data_dir)
        logger.info("Validation file download complete.")

        logger.info("Downloading test file...")
        deepchem.utils.data_utils.download_url(url=TEST_URL, dest_dir=data_dir)
        logger.info("Test file download complete")

    # Featurizing datasets
    logger.info("About to featurize UV dataset.")
    featurizer = deepchem.feat.UserDefinedFeaturizer(merck_descriptors)
    loader = deepchem.data.UserCSVLoader(tasks=UV_tasks,
                                         id_field="Molecule",
                                         featurizer=featurizer)

    logger.info("Featurizing train datasets...")
    train_dataset = loader.featurize(train_files, shard_size=shard_size)

    logger.info("Featurizing validation datasets...")
    valid_dataset = loader.featurize(valid_files, shard_size=shard_size)

    logger.info("Featurizing test datasets....")
    test_dataset = loader.featurize(test_files, shard_size=shard_size)

    # Missing entry removal
    logger.info("Removing missing entries from dataset.")
    remove_missing_entries(train_dataset)
    remove_missing_entries(valid_dataset)
    remove_missing_entries(test_dataset)

    # Shuffle the training data
    logger.info("Shuffling the training dataset")
    train_dataset.sparse_shuffle()

    # Apply transformations
    logger.info("Starting transformations")
    transformers = get_transformers(train_dataset)

    for transformer in transformers:
        logger.info("Performing transformations with {}".format(
            transformer.__class__.__name__))

        logger.info("Transforming the training dataset...")
        train_dataset = transformer.transform(train_dataset)

        logger.info("Transforming the validation dataset...")
        valid_dataset = transformer.transform(valid_dataset)

        logger.info("Transforming the test dataset...")
        test_dataset = transformer.transform(test_dataset)

    logger.info("Transformations complete.")
    logger.info("Moving datasets to corresponding directories")

    train_dataset.move(train_dir)
    logger.info("Train dataset moved.")

    valid_dataset.move(valid_dir)
    logger.info("Validation dataset moved.")

    test_dataset.move(test_dir)
    logger.info("Test dataset moved.")

    time2 = time.time()

    # TIMING
    logger.info("TIMING: UV fitting took %0.3f s" % (time2 - time1))

    return train_dataset, valid_dataset, test_dataset


def load_uv(shard_size=2000, featurizer=None, split=None, reload=True):
    """Load UV dataset; does not do train/test split

    The UV dataset is an in-house dataset from Merck that was first introduced in the following paper:
    Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.

    The UV dataset tests 10,000 of Merck's internal compounds on
    190 absorption wavelengths between 210 and 400 nm. Unlike
    most of the other datasets featured in MoleculeNet, the UV
    collection does not have structures for the compounds tested
    since they were proprietary Merck compounds. However, the
    collection does feature pre-computed descriptors for these
    compounds.

    Note that the original train/valid/test split from the source
    data was preserved here, so this function doesn't allow for
    alternate modes of splitting. Similarly, since the source data
    came pre-featurized, it is not possible to apply alternative
    featurizations.

    Parameters
    ----------
    shard_size: int, optional
        Size of the DiskDataset shards to write on disk
    featurizer: optional
        Ignored since featurization pre-computed
    split: optional
        Ignored since split pre-computed
    reload: bool, optional
        Whether to automatically re-load from disk
    """

    data_dir = deepchem.utils.data_utils.get_data_dir()
    data_dir = os.path.join(data_dir, "UV")

    if not os.path.exists(data_dir):
        os.mkdir(data_dir)

    train_dir = os.path.join(data_dir, "train_dir")
    valid_dir = os.path.join(data_dir, "valid_dir")
    test_dir = os.path.join(data_dir, "test_dir")

    if (os.path.exists(train_dir) and os.path.exists(valid_dir) and
            os.path.exists(test_dir)):

        logger.info("Reloading existing datasets")
        train_dataset = deepchem.data.DiskDataset(train_dir)
        valid_dataset = deepchem.data.DiskDataset(valid_dir)
        test_dataset = deepchem.data.DiskDataset(test_dir)

    else:
        logger.info("Featurizing datasets")
        train_dataset, valid_dataset, test_dataset = gen_uv(
            UV_tasks=UV_tasks,
            data_dir=data_dir,
            train_dir=train_dir,
            valid_dir=valid_dir,
            test_dir=test_dir,
            shard_size=shard_size)

    transformers = get_transformers(train_dataset)

    return UV_tasks, (train_dataset, valid_dataset, test_dataset), transformers



"""
Platinum Adsorbtion structure for N and NO along with their formation energies
"""
import numpy as np
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

PLATINUM_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/Platinum_adsorption.tar.gz"
PLATINUM_TASKS = ["Formation Energy"]
PRIMITIVE_CELL = {
    "lattice": [[2.818528, 0.0, 0.0], [-1.409264, 2.440917, 0.0],
                [0.0, 0.0, 25.508255]],
    "coords": [[0.66667, 0.33333, 0.090221], [0.33333, 0.66667, 0.18043936],
               [0.0, 0.0, 0.27065772], [0.66667, 0.33333, 0.36087608],
               [0.33333, 0.66667, 0.45109444], [0.0, 0.0, 0.49656991]],
    "species": ['H', 'H', 'H', 'H', 'H', 'He'],
    "site_properties": {
        'SiteTypes': ['S1', 'S1', 'S1', 'S1', 'S1', 'A1']
    }
}
PRIMITIVE_CELL_INF0 = {
    "cutoff": np.around(6.00),
    "structure": PRIMITIVE_CELL,
    "aos": ['1', '0', '2'],
    "pbc": [True, True, False],
    "ns": 1,
    "na": 1
}


class _PtAdsorptionLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, 'Platinum_adsorption.json')
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=PLATINUM_URL,
                                             dest_dir=self.data_dir)
            dc.utils.data_utils.untargz_file(
                os.path.join(self.data_dir, 'Platinum_adsorption.tar.gz'),
                self.data_dir)
        loader = dc.data.JsonLoader(tasks=PLATINUM_TASKS,
                                    feature_field="Structures",
                                    label_field="Formation Energy",
                                    featurizer=self.featurizer,
                                    **self.args)
        return loader.create_dataset(dataset_file)


def load_Platinum_Adsorption(
    featurizer: Union[dc.feat.Featurizer, str] = dc.feat.SineCoulombMatrix(),
    splitter: Union[dc.splits.Splitter, str, None] = 'random',
    transformers: List[Union[TransformerGenerator, str]] = [],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """
    Load Platinum Adsorption Dataset

    The dataset consist of diffrent configurations of Adsorbates (i.e N and NO)
    on Platinum surface represented as Lattice and their formation energy. There
    are 648 diffrent adsorbate configuration in this datasets represented as Pymatgen
    Structure objects.

    1. Pymatgen structure object with site_properties with following key value.
        - "SiteTypes", mentioning if it is a active site "A1" or spectator
            site "S1".
        - "oss", diffrent occupational sites. For spectator sites make it -1.


    Parameters
    ----------
    featurizer : Featurizer (default LCNNFeaturizer)
        the featurizer to use for processing the data. Reccomended to use
        the LCNNFeaturiser.
    splitter : Splitter (default RandomSplitter)
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data will
        be included in a single dataset.
    transformers : list of TransformerGenerators or strings. the Transformers to
        apply to the data and appropritate featuriser. Does'nt require any
        transformation for LCNN_featuriser
    reload : bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir : str
        a directory to save the raw data in
    save_dir : str, optional (default None)
        a directory to save the dataset in

    References
    ----------
    .. [1] Jonathan Lym, Geun Ho G. "Lattice Convolutional Neural Network Modeling of Adsorbate
        Coverage Effects"J. Phys. Chem. C 2019, 123, 18951−18959

    Examples
    --------
    >>>
    >> import deepchem as dc
    >> tasks, datasets, transformers = load_Platinum_Adsorption(
    >>    reload=True,
    >>    data_dir=data_path,
    >>    save_dir=data_path,
    >>    featurizer_kwargs=feat_args)
    >> train_dataset, val_dataset, test_dataset = datasets
    """

    loader = _PtAdsorptionLoader(featurizer, splitter, transformers,
                                 PLATINUM_TASKS, data_dir, save_dir, **kwargs)
    return loader.load_dataset('LCNN_feat', reload)



"""
Callback functions that can be invoked while fitting a KerasModel.
"""
import sys


class ValidationCallback(object):
    """Performs validation while training a KerasModel.

    This is a callback that can be passed to fit().  It periodically computes a
    set of metrics over a validation set, writes them to a file, and keeps track
    of the best score. In addition, it can save the best model parameters found
    so far to a directory on disk, updating them every time it finds a new best
    validation score.

    If Tensorboard logging is enabled on the KerasModel, the metrics are also
    logged to Tensorboard.  This only happens when validation coincides with a
    step on which the model writes to the log.  You should therefore make sure
    that this callback's reporting interval is an even fraction or multiple of
    the model's logging interval.
    """

    def __init__(self,
                 dataset,
                 interval,
                 metrics,
                 output_file=sys.stdout,
                 save_dir=None,
                 save_metric=0,
                 save_on_minimum=True,
                 transformers=[]):
        """Create a ValidationCallback.

        Parameters
        ----------
        dataset: dc.data.Dataset
            the validation set on which to compute the metrics
        interval: int
            the interval (in training steps) at which to perform validation
        metrics: list of dc.metrics.Metric
            metrics to compute on the validation set
        output_file: file
            to file to which results should be written
        save_dir: str
            if not None, the model parameters that produce the best validation score
            will be written to this directory
        save_metric: int
            the index of the metric to use when deciding whether to write a new set
            of parameters to disk
        save_on_minimum: bool
            if True, the best model is considered to be the one that minimizes the
            validation metric.  If False, the best model is considered to be the one
            that maximizes it.
        transformers: List[Transformer]
            List of `dc.trans.Transformer` objects. These transformations
            must have been applied to `dataset` previously. The dataset will
            be untransformed for metric evaluation.
        """
        self.dataset = dataset
        self.interval = interval
        self.metrics = metrics
        self.output_file = output_file
        self.save_dir = save_dir
        self.save_metric = save_metric
        self.save_on_minimum = save_on_minimum
        self._best_score = None
        self.transformers = transformers

    def __call__(self, model, step):
        """This is invoked by the KerasModel after every step of fitting.

        Parameters
        ----------
        model: KerasModel
            the model that is being trained
        step: int
            the index of the training step that has just completed
        """
        if step % self.interval != 0:
            return
        scores = model.evaluate(self.dataset, self.metrics, self.transformers)
        message = 'Step %d validation:' % step
        for key in scores:
            message += ' %s=%g' % (key, scores[key])
        print(message, file=self.output_file)
        if model.tensorboard:
            for key in scores:
                model._log_scalar_to_tensorboard(key, scores[key],
                                                 model.get_global_step())
        score = scores[self.metrics[self.save_metric].name]
        if not self.save_on_minimum:
            score = -score
        if self._best_score is None or score < self._best_score:
            self._best_score = score
            if self.save_dir is not None:
                model.save_checkpoint(model_dir=self.save_dir)
        if model.wandb_logger is not None:
            # Log data to Wandb
            data = {'eval/' + k: v for k, v in scores.items()}
            model.wandb_logger.log_data(data, step, dataset_id=id(self.dataset))

    def get_best_score(self):
        """This getter returns the best score evaluated on the validation set.

        Returns
        -------
        float
            The best score.
        """
        if self.save_on_minimum:
            return self._best_score
        else:
            return -self._best_score



"""
Delaney dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

DELANEY_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/delaney-processed.csv"
DELANEY_TASKS = ['measured log solubility in mols per litre']


class _DelaneyLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "delaney-processed.csv")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=DELANEY_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_delaney(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load Delaney dataset

    The Delaney (ESOL) dataset a regression dataset containing structures and
    water solubility data for 1128 compounds. The dataset is widely used to
    validate machine learning models on estimating solubility directly from
    molecular structures (as encoded in SMILES strings).

    Scaffold splitting is recommended for this dataset.

    The raw data csv file contains columns below:

    - "Compound ID" - Name of the compound
    - "smiles" - SMILES representation of the molecular structure
    - "measured log solubility in mols per litre" - Log-scale water solubility
        of the compound, used as label

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    References
    ----------
    .. [1] Delaney, John S. "ESOL: estimating aqueous solubility directly from
        molecular structure." Journal of chemical information and computer
        sciences 44.3 (2004): 1000-1005.
    """
    loader = _DelaneyLoader(featurizer, splitter, transformers, DELANEY_TASKS,
                            data_dir, save_dir, **kwargs)
    return loader.load_dataset('delaney', reload)



import pytest
try:
    import torch
except:
    pass
import numpy as np
from typing import Union


@pytest.mark.torch
def test_base_orb_params():
    from deepchem.utils.dft_utils import BaseOrbParams

    class MyOrbParams(BaseOrbParams):

        @staticmethod
        def params2orb(params, coeffs, with_penalty):
            return params, coeffs

        @staticmethod
        def orb2params(orb):
            return orb, torch.tensor([0], dtype=orb.dtype, device=orb.device)

    params = torch.randn(3, 4, 5)
    coeffs = torch.randn(3, 4, 5)
    with_penalty = 0.1
    orb, penalty = MyOrbParams.params2orb(params, coeffs, with_penalty)
    params2, coeffs2 = MyOrbParams.orb2params(orb)
    assert torch.allclose(params, params2)


@pytest.mark.torch
def test_qr_orb_params():
    from deepchem.utils.dft_utils import QROrbParams
    params = torch.randn(3, 3)
    coeffs = torch.randn(4, 3)
    with_penalty = 0.1
    orb, penalty = QROrbParams.params2orb(params, coeffs, with_penalty)
    params2, coeffs2 = QROrbParams.orb2params(orb)
    assert torch.allclose(orb, params2)


@pytest.mark.torch
def test_mat_exp_orb_params():
    from deepchem.utils.dft_utils import MatExpOrbParams
    params = torch.randn(3, 3)
    coeffs = torch.randn(4, 3)
    orb = MatExpOrbParams.params2orb(params, coeffs)[0]
    params2, coeffs2 = MatExpOrbParams.orb2params(orb)
    assert coeffs2.shape == orb.shape


@pytest.mark.torch
def test_lattice():
    """Test lattice object.
    Comparing it's output with it's original implementation in dqc.
    """
    from deepchem.utils.dft_utils import Lattice
    a = torch.tensor([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
    lattice = Lattice(a)

    assert torch.allclose(lattice.lattice_vectors(), a)
    assert torch.allclose(lattice.recip_vectors(),
                          torch.inverse(a.transpose(-2, -1)) * (2 * np.pi))
    assert torch.allclose(lattice.volume(), torch.det(a))

    assert torch.allclose(
        lattice.get_lattice_ls(1.0),
        torch.tensor([[0., 0., -1.], [0., -1., 0.], [-1., 0., 0.], [0., 0., 0.],
                      [1., 0., 0.], [0., 1., 0.], [0., 0., 1.]]))
    assert torch.allclose(
        lattice.get_gvgrids(6.0)[0],
        torch.tensor([[0.0000, 0.0000, -6.2832], [0.0000, -6.2832, 0.0000],
                      [-6.2832, 0.0000, 0.0000], [0.0000, 0.0000, 0.0000],
                      [6.2832, 0.0000, 0.0000], [0.0000, 6.2832, 0.0000],
                      [0.0000, 0.0000, 6.2832]]))
    assert torch.allclose(
        lattice.get_gvgrids(6.0)[1],
        torch.tensor([1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000, 1.0000]))
    assert lattice.estimate_ewald_eta(1e-5) == 1.8


@pytest.mark.torch
def test_parse_moldesc():
    """Tests Moldesc Parser."""
    from deepchem.utils.dft_utils import parse_moldesc
    system = {
        'type': 'mol',
        'kwargs': {
            'moldesc': 'H 0.86625 0 0; F -0.86625 0 0',
            'basis': '6-311++G(3df,3pd)'
        }
    }
    atomzs, atomposs = parse_moldesc(system["kwargs"]["moldesc"])
    assert torch.allclose(atomzs, torch.tensor([1., 9.], dtype=torch.float64))
    assert torch.allclose(
        atomposs,
        torch.tensor(
            [[0.86625, 0.00000, 0.00000], [-0.86625, 0.00000, 0.00000]],
            dtype=torch.float64))
    system2 = (['H', 'F'], torch.tensor([[0.86625, 0, 0], [-0.86625, 0, 0]]))
    atomzs2, atomposs2 = parse_moldesc(system2)
    assert torch.allclose(atomzs2, torch.tensor([1., 9.], dtype=torch.float64))
    assert torch.allclose(
        atomposs2,
        torch.tensor(
            [[0.86625, 0.00000, 0.00000], [-0.86625, 0.00000, 0.00000]],
            dtype=torch.float64))


@pytest.mark.torch
def test_spin_param():
    """Test SpinParam object."""
    from deepchem.utils.dft_utils import SpinParam
    dens_u = torch.ones(1)
    dens_d = torch.zeros(1)
    sp = SpinParam(u=dens_u, d=dens_d)

    assert torch.allclose(sp.u, dens_u)
    assert torch.allclose(sp.d, dens_d)
    assert torch.allclose(sp.sum(), torch.tensor([1.]))
    assert torch.allclose(sp.reduce(torch.multiply), torch.tensor([0.]))


@pytest.mark.torch
def test_val_grad():
    """Test ValGrad object."""
    from deepchem.utils.dft_utils import ValGrad
    dens = torch.ones(1)
    grad = torch.zeros(1)
    lapl = torch.ones(1)
    kin = torch.ones(1)
    vg = ValGrad(value=dens, grad=grad, lapl=lapl, kin=kin)

    assert torch.allclose(vg.value, dens)
    assert torch.allclose(vg.grad, grad)
    assert torch.allclose(vg.lapl, lapl)
    assert torch.allclose(vg.kin, kin)

    vg2 = vg + vg
    assert torch.allclose(vg2.value, torch.tensor([2.]))
    assert torch.allclose(vg2.grad, torch.tensor([0.]))
    assert torch.allclose(vg2.lapl, torch.tensor([2.]))
    assert torch.allclose(vg2.kin, torch.tensor([2.]))

    vg5 = vg * 5
    assert torch.allclose(vg5.value, torch.tensor([5.]))
    assert torch.allclose(vg5.grad, torch.tensor([0.]))
    assert torch.allclose(vg5.lapl, torch.tensor([5.]))
    assert torch.allclose(vg5.kin, torch.tensor([5.]))


@pytest.mark.torch
def test_base_xc():
    """Test BaseXC."""
    from deepchem.utils.dft_utils import ValGrad, SpinParam
    from deepchem.utils.dft_utils import BaseXC

    class MyXC(BaseXC):

        @property
        def family(self) -> int:
            return 1

        def get_edensityxc(
                self, densinfo: Union[ValGrad,
                                      SpinParam[ValGrad]]) -> torch.Tensor:
            if isinstance(densinfo, ValGrad):
                return densinfo.value.pow(2)
            else:
                return densinfo.u.value.pow(2) + densinfo.d.value.pow(2)

        def get_vxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]
        ) -> Union[ValGrad, SpinParam[ValGrad]]:
            if isinstance(densinfo, ValGrad):
                return ValGrad(value=2 * densinfo.value)
            else:
                return SpinParam(u=ValGrad(value=2 * densinfo.u.value),
                                 d=ValGrad(value=2 * densinfo.d.value))

    xc = MyXC()
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc.get_edensityxc(densinfo_v),
                          torch.tensor([1., 4., 9.]))
    assert xc.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([2., 4., 6.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc.get_edensityxc(densinfo_s),
                          torch.tensor([17., 29., 45.]))
    assert xc.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([2., 4., 6.])),
        d=ValGrad(value=torch.tensor([8., 10., 12.])))


@pytest.mark.torch
def test_addbase_xc():
    """Test AddBaseXC."""
    from deepchem.utils.dft_utils import ValGrad, SpinParam
    from deepchem.utils.dft_utils import BaseXC, AddBaseXC

    class MyXC(BaseXC):

        @property
        def family(self) -> int:
            return 1

        def get_edensityxc(
                self, densinfo: Union[ValGrad,
                                      SpinParam[ValGrad]]) -> torch.Tensor:
            if isinstance(densinfo, ValGrad):
                return densinfo.value.pow(2)
            else:
                return densinfo.u.value.pow(2) + densinfo.d.value.pow(2)

        def get_vxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]
        ) -> Union[ValGrad, SpinParam[ValGrad]]:
            if isinstance(densinfo, ValGrad):
                return ValGrad(value=2 * densinfo.value)
            else:
                return SpinParam(u=ValGrad(value=2 * densinfo.u.value),
                                 d=ValGrad(value=2 * densinfo.d.value))

    xc1 = MyXC()
    xc = xc1 + xc1
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))

    xc2 = AddBaseXC(xc1, xc1)
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc2.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc2.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc2.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc2.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))


@pytest.mark.torch
def test_mulbase_xc():
    """Test AddBaseXC."""
    from deepchem.utils.dft_utils import ValGrad, SpinParam
    from deepchem.utils.dft_utils import BaseXC, MulBaseXC

    class MyXC(BaseXC):

        @property
        def family(self) -> int:
            return 1

        def get_edensityxc(
                self, densinfo: Union[ValGrad,
                                      SpinParam[ValGrad]]) -> torch.Tensor:
            if isinstance(densinfo, ValGrad):
                return densinfo.value.pow(2)
            else:
                return densinfo.u.value.pow(2) + densinfo.d.value.pow(2)

        def get_vxc(
            self, densinfo: Union[ValGrad, SpinParam[ValGrad]]
        ) -> Union[ValGrad, SpinParam[ValGrad]]:
            if isinstance(densinfo, ValGrad):
                return ValGrad(value=2 * densinfo.value)
            else:
                return SpinParam(u=ValGrad(value=2 * densinfo.u.value),
                                 d=ValGrad(value=2 * densinfo.d.value))

    xc1 = MyXC()
    xc = xc1 * 2
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))

    xc2 = MulBaseXC(xc1, 2)
    densinfo_v = ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True))
    assert torch.allclose(xc2.get_edensityxc(densinfo_v),
                          torch.tensor([2., 8., 18.]))
    assert xc2.get_vxc(densinfo_v) == ValGrad(value=torch.tensor([4., 8., 12.]))

    densinfo_s = SpinParam(
        u=ValGrad(value=torch.tensor([1., 2., 3.], requires_grad=True)),
        d=ValGrad(value=torch.tensor([4., 5., 6.], requires_grad=True)))
    assert torch.allclose(xc2.get_edensityxc(densinfo_s),
                          torch.tensor([34., 58., 90.]))
    assert xc2.get_vxc(densinfo_s) == SpinParam(
        u=ValGrad(value=torch.tensor([4., 8., 12.])),
        d=ValGrad(value=torch.tensor([16., 20., 24.])))


@pytest.mark.torch
def test_base_grid():
    import torch
    from deepchem.utils.dft_utils import BaseGrid

    class Grid(BaseGrid):

        def __init__(self):
            super(Grid, self).__init__()
            self.ngrid = 10
            self.ndim = 3
            self.dvolume = torch.ones(self.ngrid)
            self.rgrid = torch.ones((self.ngrid, self.ndim))

        def get_dvolume(self):
            return self.dvolume

        def get_rgrid(self):
            return self.rgrid

    grid = Grid()
    assert torch.allclose(grid.get_dvolume(), torch.ones(10))
    assert torch.allclose(grid.get_rgrid(), torch.ones((10, 3)))


@pytest.mark.torch
def test_base_df():
    """Test BaseDF. Checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils import BaseDF

    class MyDF(BaseDF):

        def __init__(self):
            super(MyDF, self).__init__()

        def get_j2c(self):
            return torch.ones((3, 3))

        def get_j3c(self):
            return torch.ones((3, 3, 3))

    df = MyDF()
    assert torch.allclose(df.get_j2c(), torch.ones((3, 3)))
    assert torch.allclose(df.get_j3c(), torch.ones((3, 3, 3)))


@pytest.mark.torch
def test_base_hamilton():
    """Test BaseHamilton. Checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils import BaseHamilton

    class MyHamilton(BaseHamilton):

        def __init__(self):
            self._nao = 2
            self._kpts = torch.tensor([[0.0, 0.0, 0.0]])
            self._df = None

        @property
        def nao(self):
            return self._nao

        @property
        def kpts(self):
            return self._kpts

        @property
        def df(self):
            return self._df

        def build(self):
            return self

        def get_nuclattr(self):
            return torch.ones((1, 1, self.nao, self.nao))

    ham = MyHamilton()
    hamilton = ham.build()
    assert torch.allclose(hamilton.get_nuclattr(),
                          torch.tensor([[[[1., 1.], [1., 1.]]]]))


@pytest.mark.torch
def test_cgto_basis():
    from deepchem.utils.dft_utils import CGTOBasis
    alphas = torch.ones(1)
    coeffs = torch.ones(1)
    cgto = CGTOBasis(angmom=0, alphas=alphas, coeffs=coeffs)
    cgto.wfnormalize_()

    if cgto.normalized is True:
        assert True
    else:
        assert False


@pytest.mark.torch
def test_atom_cgto_basis():
    from deepchem.utils.dft_utils import AtomCGTOBasis, CGTOBasis
    alphas = torch.ones(1)
    coeffs = torch.ones(1)
    cgto = CGTOBasis(angmom=0, alphas=alphas, coeffs=coeffs)
    atomcgto = AtomCGTOBasis(atomz=1, bases=[cgto], pos=[0.0, 0.0, 0.0])
    assert atomcgto.bases[0] == cgto


@pytest.mark.torch
def test_base_system():
    """Test BaseSystem. Checks that it doesn't raise errors."""
    from deepchem.utils.dft_utils import BaseSystem, BaseHamilton, BaseGrid

    class MySystem(BaseSystem):

        def __init__(self):
            self.hamiltonian = BaseHamilton()
            self.grid = BaseGrid()

        def get_hamiltonian(self):
            return self.hamiltonian

        def get_grid(self):
            return self.grid

        def requires_grid(self):
            return True

    system = MySystem()
    assert system.requires_grid()


@pytest.mark.torch
def test_radial_grid():
    from deepchem.utils.dft_utils import RadialGrid
    grid = RadialGrid(4, grid_integrator="chebyshev", grid_transform="logm3")
    assert grid.get_rgrid().shape == torch.Size([4, 1])
    assert grid.get_dvolume().shape == torch.Size([4])


@pytest.mark.torch
def test_get_xw_integration():
    from deepchem.utils.dft_utils import get_xw_integration
    x, w = get_xw_integration(4, "chebyshev")
    assert x.shape == (4,)
    assert w.shape == torch.Size([4])


@pytest.mark.torch
def test_sliced_radial_grid():
    from deepchem.utils.dft_utils import RadialGrid, SlicedRadialGrid
    grid = RadialGrid(4)
    sliced_grid = SlicedRadialGrid(grid, 2)
    assert sliced_grid.get_rgrid().shape == torch.Size([1])


@pytest.mark.torch
def test_de2_transform():
    from deepchem.utils.dft_utils import DE2Transformation
    x = torch.linspace(-1, 1, 100)
    r = DE2Transformation().x2r(x)
    assert r.shape == torch.Size([100])
    drdx = DE2Transformation().get_drdx(x)
    assert drdx.shape == torch.Size([100])


@pytest.mark.torch
def test_logm3_transform():
    from deepchem.utils.dft_utils import LogM3Transformation
    x = torch.linspace(-1, 1, 100)
    r = LogM3Transformation().x2r(x)
    assert r.shape == torch.Size([100])
    drdx = LogM3Transformation().get_drdx(x)
    assert drdx.shape == torch.Size([100])


@pytest.mark.torch
def test_treutlerm4_transform():
    from deepchem.utils.dft_utils import TreutlerM4Transformation
    x = torch.linspace(-1, 1, 100)
    r = TreutlerM4Transformation().x2r(x)
    assert r.shape == torch.Size([100])
    drdx = TreutlerM4Transformation().get_drdx(x)
    assert drdx.shape == torch.Size([100])


@pytest.mark.torch
def test_get_grid_transform():
    from deepchem.utils.dft_utils import get_grid_transform
    transform = get_grid_transform("logm3")
    transform.x2r(torch.tensor([0.5])) == torch.tensor([2.])


@pytest.mark.torch
def test_SCF_QCCalc():
    from deepchem.utils.dft_utils import SCF_QCCalc, BaseSCFEngine, SpinParam

    # Define the engine
    class engine(BaseSCFEngine):

        def polarised():
            return False

        def dm2energy(
                self, dm: Union[torch.Tensor,
                                SpinParam[torch.Tensor]]) -> torch.Tensor:
            if isinstance(dm, SpinParam):
                return dm.u + dm.d * 1.1
            return dm * 1.1

    myEngine = engine()
    a = SCF_QCCalc(myEngine)
    assert torch.allclose(a.dm2energy(torch.tensor([1.1])),
                          torch.tensor([1.2100]))


@pytest.mark.torch
def test_BaseSCFEngine():
    from deepchem.utils.dft_utils import BaseSCFEngine, SpinParam

    class engine(BaseSCFEngine):

        def polarised():
            return False

        def dm2energy(
                self, dm: Union[torch.Tensor,
                                SpinParam[torch.Tensor]]) -> torch.Tensor:
            if isinstance(dm, SpinParam):
                return dm.u + dm.d * 1.1
            return dm * 1.1

    myEngine = engine()
    assert myEngine.dm2energy(torch.tensor(1.2)) == torch.tensor(1.32)


@pytest.mark.torch
def test_hf_engine():
    """Tests HFEngine and methods of its parent class BaseSCFEngine."""
    from deepchem.utils.dft_utils import (BaseHamilton, BaseSystem, BaseGrid,
                                          SpinParam, HFEngine)
    from deepchem.utils.differentiation_utils import LinearOperator

    class MyLinOp(LinearOperator):

        def __init__(self, shape):
            super(MyLinOp, self).__init__(shape)
            self.param = torch.rand(shape)

        def _getparamnames(self, prefix=""):
            return [prefix + "param"]

        def _mv(self, x):
            return torch.matmul(self.param, x)

        def _rmv(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _mm(self, x):
            return torch.matmul(self.param, x)

        def _rmm(self, x):
            return torch.matmul(self.param.transpose(-2, -1).conj(), x)

        def _fullmatrix(self):
            return self.param

    class MyHamilton(BaseHamilton):

        def __init__(self):
            self._nao = 2
            self._kpts = torch.tensor([[0.0, 0.0, 0.0]])
            self._df = None

        @property
        def nao(self):
            return self._nao

        @property
        def kpts(self):
            return self._kpts

        @property
        def df(self):
            return self._df

        def build(self):
            return self

        def get_nuclattr(self):
            return torch.ones((1, 1, self.nao, self.nao))

        def get_e_elrep(self, dmtot):
            return 2 * dmtot

        def get_e_exchange(self, dm):
            return 3 * dm

        def get_e_hcore(self, dm):
            return 4.0 * dm

        def get_elrep(self, dmtot):
            return MyLinOp((self.nao + 1, self.nao + 1))

        def get_exchange(self, dm):
            return MyLinOp((self.nao + 1, self.nao + 1))

        def get_kinnucl(self):
            linop = MyLinOp((self.nao + 1, self.nao + 1))
            return linop

        def ao_orb2dm(self, orb: torch.Tensor,
                      orb_weight: torch.Tensor) -> torch.Tensor:
            return orb * orb_weight

    ham = MyHamilton()

    class MySystem(BaseSystem):

        def __init__(self):
            self.hamiltonian = ham
            self.grid = BaseGrid()

        def get_hamiltonian(self):
            return self.hamiltonian

        def get_grid(self):
            return self.grid

        def requires_grid(self):
            return True

        def get_orbweight(self, polarized: bool = False):
            return SpinParam(torch.tensor([1.0]), torch.tensor([2.0]))

        def get_nuclei_energy(self):
            return 10.0

    system = MySystem()
    engine = HFEngine(system)
    engine.set_eigen_options(eigen_options={"method": "exacteig"})

    assert engine.dm2energy(torch.tensor([2])) == torch.tensor([28.0])
    assert engine.dm2scp(torch.tensor([2])).shape == torch.Size([3, 3])
    assert engine.scp2dm(torch.rand((2, 2, 2))).u.shape == torch.Size([2, 1])



"""
qm9 dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

GDB9_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb9.tar.gz"
QM9_CSV_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/qm9.csv"
QM9_TASKS = [
    "mu", "alpha", "homo", "lumo", "gap", "r2", "zpve", "cv", "u0", "u298",
    "h298", "g298"
]


class _QM9Loader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "gdb9.sdf")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=GDB9_URL,
                                             dest_dir=self.data_dir)
            dc.utils.data_utils.untargz_file(
                os.path.join(self.data_dir, "gdb9.tar.gz"), self.data_dir)
        loader = dc.data.SDFLoader(tasks=self.tasks,
                                   featurizer=self.featurizer,
                                   sanitize=True)
        return loader.create_dataset(dataset_file, shard_size=4096)


def load_qm9(
    featurizer: Union[dc.feat.Featurizer, str] = dc.feat.CoulombMatrix(29),
    splitter: Union[dc.splits.Splitter, str, None] = 'random',
    transformers: List[Union[TransformerGenerator, str]] = ['normalization'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load QM9 dataset

    QM9 is a comprehensive dataset that provides geometric, energetic,
    electronic and thermodynamic properties for a subset of GDB-17
    database, comprising 134 thousand stable organic molecules with up
    to 9 heavy atoms.  All molecules are modeled using density
    functional theory (B3LYP/6-31G(2df,p) based DFT).

    Random splitting is recommended for this dataset.

    The source data contain:

    - qm9.sdf: molecular structures
    - qm9.sdf.csv: tables for molecular properties

    - "mol_id" - Molecule ID (gdb9 index) mapping to the .sdf file
    - "A" - Rotational constant (unit: GHz)
    - "B" - Rotational constant (unit: GHz)
    - "C" - Rotational constant (unit: GHz)
    - "mu" - Dipole moment (unit: D)
    - "alpha" - Isotropic polarizability (unit: Bohr^3)
    - "homo" - Highest occupied molecular orbital energy (unit: Hartree)
    - "lumo" - Lowest unoccupied molecular orbital energy (unit: Hartree)
    - "gap" - Gap between HOMO and LUMO (unit: Hartree)
    - "r2" - Electronic spatial extent (unit: Bohr^2)
    - "zpve" - Zero point vibrational energy (unit: Hartree)
    - "u0" - Internal energy at 0K (unit: Hartree)
    - "u298" - Internal energy at 298.15K (unit: Hartree)
    - "h298" - Enthalpy at 298.15K (unit: Hartree)
    - "g298" - Free energy at 298.15K (unit: Hartree)
    - "cv" - Heat capavity at 298.15K (unit: cal/(mol*K))
    - "u0_atom" - Atomization energy at 0K (unit: kcal/mol)
    - "u298_atom" - Atomization energy at 298.15K (unit: kcal/mol)
    - "h298_atom" - Atomization enthalpy at 298.15K (unit: kcal/mol)
    - "g298_atom" - Atomization free energy at 298.15K (unit: kcal/mol)

    "u0_atom" ~ "g298_atom" (used in MoleculeNet) are calculated from the
    differences between "u0" ~ "g298" and sum of reference energies of all
    atoms in the molecules, as given in
    https://figshare.com/articles/Atomref%3A_Reference_thermochemical_energies_of_H%2C_C%2C_N%2C_O%2C_F_atoms./1057643

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    Note
    ----
    DeepChem 2.4.0 has turned on sanitization for this dataset by
    default.  For the QM9 dataset, this means that calling this
    function will return 132480 compounds instead of 133885 in the
    source dataset file. This appears to be due to valence
    specification mismatches in the dataset that weren't caught in
    earlier more lax versions of RDKit. Note that this may subtly
    affect benchmarking results on this dataset.

    References
    ----------
    .. [1] Blum, Lorenz C., and Jean-Louis Reymond. "970 million druglike small
        molecules for virtual screening in the chemical universe database GDB-13."
        Journal of the American Chemical Society 131.25 (2009): 8732-8733.
    .. [2] Ramakrishnan, Raghunathan, et al. "Quantum chemistry structures and
        properties of 134 kilo molecules." Scientific data 1 (2014): 140022.
    """
    loader = _QM9Loader(featurizer, splitter, transformers, QM9_TASKS, data_dir,
                        save_dir, **kwargs)
    return loader.load_dataset('qm9', reload)



"""
This is a sample implementation for working DGL with DeepChem!
"""
import torch
import torch.nn as nn
import torch.nn.functional as F

from deepchem.models.losses import Loss, L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.torch_models.torch_model import TorchModel


class CGCNNLayer(nn.Module):
    """The convolutional layer of CGCNN.

    This class was implemented using DGLGraph methods.
    Please confirm how to use DGLGraph methods from below link.
    See: https://docs.dgl.ai/en/0.4.x/tutorials/models/1_gnn/9_gat.html

    Examples
    --------
    >>> import deepchem as dc
    >>> from pymatgen.core import Lattice, Structure
    >>> lattice = Lattice.cubic(4.2)
    >>> structure = Structure(lattice, ["Cs", "Cl"], [[0, 0, 0], [0.5, 0.5, 0.5]])
    >>> featurizer = dc.feat.CGCNNFeaturizer()
    >>> cgcnn_graph = featurizer.featurize([structure])[0]
    >>> cgcnn_graph.num_node_features
    92
    >>> cgcnn_graph.num_edge_features
    41
    >>> cgcnn_dgl_graph = cgcnn_graph.to_dgl_graph()
    >>> print(type(cgcnn_dgl_graph))
    <class 'dgl.heterograph.DGLGraph'>
    >>> layer = CGCNNLayer(hidden_node_dim=92, edge_dim=41)
    >>> node_feats = cgcnn_dgl_graph.ndata.pop('x')
    >>> edge_feats = cgcnn_dgl_graph.edata.pop('edge_attr')
    >>> new_node_feats, new_edge_feats = layer(cgcnn_dgl_graph, node_feats, edge_feats)

    Notes
    -----
    This class requires DGL and PyTorch to be installed.
    """

    def __init__(self,
                 hidden_node_dim: int,
                 edge_dim: int,
                 batch_norm: bool = True):
        """
        Parameters
        ----------
        hidden_node_dim: int
            The length of the hidden node feature vectors.
        edge_dim: int
            The length of the edge feature vectors.
        batch_norm: bool, default True
            Whether to apply batch normalization or not.
        """
        super(CGCNNLayer, self).__init__()
        z_dim = 2 * hidden_node_dim + edge_dim
        liner_out_dim = 2 * hidden_node_dim
        self.linear = nn.Linear(z_dim, liner_out_dim)
        self.batch_norm = nn.BatchNorm1d(liner_out_dim) if batch_norm else None

    def message_func(self, edges):
        z = torch.cat([edges.src['x'], edges.dst['x'], edges.data['edge_attr']],
                      dim=1)
        z = self.linear(z)
        if self.batch_norm is not None:
            z = self.batch_norm(z)
        gated_z, message_z = z.chunk(2, dim=1)
        gated_z = torch.sigmoid(gated_z)
        message_z = F.softplus(message_z)
        return {'message': gated_z * message_z}

    def reduce_func(self, nodes):
        nbr_sumed = torch.sum(nodes.mailbox['message'], dim=1)
        new_x = F.softplus(nodes.data['x'] + nbr_sumed)
        return {'new_x': new_x}

    def forward(self, dgl_graph, node_feats, edge_feats):
        """Update node representations.

        Parameters
        ----------
        dgl_graph: DGLGraph
            DGLGraph for a batch of graphs.
        node_feats: torch.Tensor
            The node features. The shape is `(N, hidden_node_dim)`.
        edge_feats: torch.Tensor
            The edge features. The shape is `(N, hidden_node_dim)`.

        Returns
        -------
        node_feats: torch.Tensor
            The updated node features. The shape is `(N, hidden_node_dim)`.
        """
        dgl_graph.ndata['x'] = node_feats
        dgl_graph.edata['edge_attr'] = edge_feats
        dgl_graph.update_all(self.message_func, self.reduce_func)
        node_feats = dgl_graph.ndata.pop('new_x')
        return node_feats


class CGCNN(nn.Module):
    """Crystal Graph Convolutional Neural Network (CGCNN).

    This model takes arbitary crystal structures as an input, and predict material properties
    using the element information and connection of atoms in the crystal. If you want to get
    some material properties which has a high computational cost like band gap in the case
    of DFT, this model may be useful. This model is one of variants of Graph Convolutional
    Networks. The main differences between other GCN models are how to construct graphs and
    how to update node representations. This model defines the crystal graph from structures
    using distances between atoms. The crystal graph is an undirected multigraph which is defined
    by nodes representing atom properties and edges representing connections between atoms
    in a crystal. And, this model updates the node representations using both neighbor node
    and edge representations. Please confirm the detail algorithms from [1]_.

    Examples
    --------
    >>> import deepchem as dc
    >>> from pymatgen.core import Lattice, Structure
    >>> lattice = Lattice.cubic(4.2)
    >>> structure = Structure(lattice, ["Cs", "Cl"], [[0, 0, 0], [0.5, 0.5, 0.5]])
    >>> featurizer = dc.feat.CGCNNFeaturizer()
    >>> cgcnn_feat = featurizer.featurize([structure])[0]
    >>> print(type(cgcnn_feat))
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> cgcnn_dgl_feat = cgcnn_feat.to_dgl_graph()
    >>> print(type(cgcnn_dgl_feat))
    <class 'dgl.heterograph.DGLGraph'>
    >>> model = dc.models.CGCNN(mode='regression', n_tasks=2)
    >>> out = model(cgcnn_dgl_feat)
    >>> print(type(out))
    <class 'torch.Tensor'>
    >>> out.shape == (1, 2)
    True

    References
    ----------
    .. [1] Xie, Tian, and Jeffrey C. Grossman. "Crystal graph convolutional neural networks
        for an accurate and interpretable prediction of material properties." Physical review letters
        120.14 (2018): 145301.

    Notes
    -----
    This class requires DGL and PyTorch to be installed.
    """

    def __init__(
        self,
        in_node_dim: int = 92,
        hidden_node_dim: int = 64,
        in_edge_dim: int = 41,
        num_conv: int = 3,
        predictor_hidden_feats: int = 128,
        n_tasks: int = 1,
        mode: str = 'regression',
        n_classes: int = 2,
    ):
        """
        Parameters
        ----------
        in_node_dim: int, default 92
            The length of the initial node feature vectors. The 92 is
            based on length of vectors in the atom_init.json.
        hidden_node_dim: int, default 64
            The length of the hidden node feature vectors.
        in_edge_dim: int, default 41
            The length of the initial edge feature vectors. The 41 is
            based on default setting of CGCNNFeaturizer.
        num_conv: int, default 3
            The number of convolutional layers.
        predictor_hidden_feats: int, default 128
            The size for hidden representations in the output MLP predictor.
        n_tasks: int, default 1
            The number of the output size.
        mode: str, default 'regression'
            The model type, 'classification' or 'regression'.
        n_classes: int, default 2
            The number of classes to predict (only used in classification mode).
        """
        try:
            import dgl
        except:
            raise ImportError("This class requires DGL to be installed.")
        super(CGCNN, self).__init__()
        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        self.n_tasks = n_tasks
        self.mode = mode
        self.n_classes = n_classes
        self.embedding = nn.Linear(in_node_dim, hidden_node_dim)
        self.conv_layers = nn.ModuleList([
            CGCNNLayer(hidden_node_dim=hidden_node_dim,
                       edge_dim=in_edge_dim,
                       batch_norm=True) for _ in range(num_conv)
        ])
        self.pooling = dgl.mean_nodes
        self.fc = nn.Linear(hidden_node_dim, predictor_hidden_feats)
        if self.mode == 'regression':
            self.out = nn.Linear(predictor_hidden_feats, n_tasks)
        else:
            self.out = nn.Linear(predictor_hidden_feats, n_tasks * n_classes)

    def forward(self, dgl_graph):
        """Predict labels

        Parameters
        ----------
        dgl_graph: DGLGraph
            DGLGraph for a batch of graphs. The graph expects that the node features
            are stored in `ndata['x']`, and the edge features are stored in `edata['edge_attr']`.

        Returns
        -------
        out: torch.Tensor
            The output values of this model.
            If mode == 'regression', the shape is `(batch_size, n_tasks)`.
            If mode == 'classification', the shape is `(batch_size, n_tasks, n_classes)` (n_tasks > 1)
            or `(batch_size, n_classes)` (n_tasks == 1) and the output values are probabilities of each class label.
        """
        graph = dgl_graph
        # embedding node features
        node_feats = graph.ndata.pop('x')
        edge_feats = graph.edata.pop('edge_attr')
        node_feats = self.embedding(node_feats)

        # convolutional layer
        for conv in self.conv_layers:
            node_feats = conv(graph, node_feats, edge_feats)

        # pooling
        graph.ndata['updated_x'] = node_feats
        graph_feat = F.softplus(self.pooling(graph, 'updated_x'))
        graph_feat = F.softplus(self.fc(graph_feat))
        out = self.out(graph_feat)

        if self.mode == 'regression':
            return out
        else:
            logits = out.view(-1, self.n_tasks, self.n_classes)
            # for n_tasks == 1 case
            logits = torch.squeeze(logits)
            proba = F.softmax(logits)
            return proba, logits


class CGCNNModel(TorchModel):
    """Crystal Graph Convolutional Neural Network (CGCNN).

    Here is a simple example of code that uses the CGCNNModel with
    materials dataset.

    Examples
    --------
    >>> import deepchem as dc
    >>> dataset_config = {"reload": False, "featurizer": dc.feat.CGCNNFeaturizer(), "transformers": []}
    >>> tasks, datasets, transformers = dc.molnet.load_perovskite(**dataset_config)
    >>> train, valid, test = datasets
    >>> model = dc.models.CGCNNModel(mode='regression', batch_size=32, learning_rate=0.001)
    >>> avg_loss = model.fit(train, nb_epoch=50)

    This model takes arbitary crystal structures as an input, and predict material properties
    using the element information and connection of atoms in the crystal. If you want to get
    some material properties which has a high computational cost like band gap in the case
    of DFT, this model may be useful. This model is one of variants of Graph Convolutional
    Networks. The main differences between other GCN models are how to construct graphs and
    how to update node representations. This model defines the crystal graph from structures
    using distances between atoms. The crystal graph is an undirected multigraph which is defined
    by nodes representing atom properties and edges representing connections between atoms
    in a crystal. And, this model updates the node representations using both neighbor node
    and edge representations. Please confirm the detail algorithms from [1]_.

    References
    ----------
    .. [1] Xie, Tian, and Jeffrey C. Grossman. "Crystal graph convolutional neural networks
        for an accurate and interpretable prediction of material properties." Physical review letters
        120.14 (2018): 145301.

    Notes
    -----
    This class requires DGL and PyTorch to be installed.
    """

    def __init__(self,
                 in_node_dim: int = 92,
                 hidden_node_dim: int = 64,
                 in_edge_dim: int = 41,
                 num_conv: int = 3,
                 predictor_hidden_feats: int = 128,
                 n_tasks: int = 1,
                 mode: str = 'regression',
                 n_classes: int = 2,
                 **kwargs):
        """This class accepts all the keyword arguments from TorchModel.

        Parameters
        ----------
        in_node_dim: int, default 92
            The length of the initial node feature vectors. The 92 is
            based on length of vectors in the atom_init.json.
        hidden_node_dim: int, default 64
            The length of the hidden node feature vectors.
        in_edge_dim: int, default 41
            The length of the initial edge feature vectors. The 41 is
            based on default setting of CGCNNFeaturizer.
        num_conv: int, default 3
            The number of convolutional layers.
        predictor_hidden_feats: int, default 128
            The size for hidden representations in the output MLP predictor.
        n_tasks: int, default 1
            The number of the output size.
        mode: str, default 'regression'
            The model type, 'classification' or 'regression'.
        n_classes: int, default 2
            The number of classes to predict (only used in classification mode).
        kwargs: Dict
            This class accepts all the keyword arguments from TorchModel.
        """
        model = CGCNN(in_node_dim, hidden_node_dim, in_edge_dim, num_conv,
                      predictor_hidden_feats, n_tasks, mode, n_classes)
        if mode == "regression":
            loss: Loss = L2Loss()
            output_types = ['prediction']
        else:
            loss = SparseSoftmaxCrossEntropy()
            output_types = ['prediction', 'loss']
        super(CGCNNModel, self).__init__(model,
                                         loss=loss,
                                         output_types=output_types,
                                         **kwargs)

    def _prepare_batch(self, batch):
        """Create batch data for CGCNN.

        Parameters
        ----------
        batch: Tuple
            The tuple are `(inputs, labels, weights)`.

        Returns
        -------
        inputs: DGLGraph
            DGLGraph for a batch of graphs.
        labels: List[torch.Tensor] or None
            The labels converted to torch.Tensor
        weights: List[torch.Tensor] or None
            The weights for each sample or sample/task pair converted to torch.Tensor
        """
        try:
            import dgl
        except:
            raise ImportError("This class requires DGL to be installed.")

        inputs, labels, weights = batch
        dgl_graphs = [graph.to_dgl_graph() for graph in inputs[0]]
        inputs = dgl.batch(dgl_graphs).to(self.device)
        _, labels, weights = super(CGCNNModel, self)._prepare_batch(
            ([], labels, weights))
        return inputs, labels, weights
